import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

/*
 * Created on Apr 5, 2005 by Sam Danziger
 *
 * 
 */
public class Bayes {
    /**
     * Internal Class that holds the name,tp,fp,tn,fn,ppv,and npv for a 
     * Component Classifier
     * @author sdanzige
     *
     */
	class ComponentClassifier implements Comparable{
		String name;
		Integer tp;
		Integer fp;
		Integer fn;
		Integer tn;
		Double ppv; // tp/(tp+fp)
		Double npv; // tn/(tn+fn)
		Double basePositive; // (tp+fn)/N
		Double baseNegative; // (tn+fp)/N
		Double modPPV; //ppv - basePositive
		Double modNPV; //npv - baseNegative
		
    	/**
    	 * Create a new component classifier
    	 * @param _name
    	 * @param _tp
    	 * @param _fp
    	 * @param _fn
    	 * @param _tn
    	 */
		public ComponentClassifier(String _name, int _tp, int _fp, int _fn, int _tn) {
    		name=_name;
    		tp=new Integer(_tp);
    		fp=new Integer(_fp);
    		fn=new Integer(_fn);
    		tn=new Integer(_tn);
    		
    		//Raw positive predictive value
    		double dppv=(double)_tp/((double)_tp+(double)_fp);
    		ppv=new Double(dppv);
    		double dnpv=(double)_tn/((double)_tn+(double)_fn);
    		npv=new Double(dnpv);
    		
    		//Baseline chance of accuracy
    		double N=(double)_tp+(double)_fp+(double)_fn+(double)_tn;
    		double dbasePositive=((double)_tp+(double)_fn) / N;
    		basePositive = new Double (dbasePositive);
    		double dbaseNegative=((double)_tn+(double)_fp) / N;
    		baseNegative = new Double (dbaseNegative);
    		
    		//The modified positive and negative predictive values;
    		double dmodPPV=dppv-dbasePositive;
    		if (dmodPPV<0) {dmodPPV=0;}
    		modPPV=new Double(dmodPPV);
    		
    		double dmodNPV=dnpv-dbaseNegative;
    		if (dmodNPV<0) {dmodNPV=0;}
    		modNPV=new Double(dmodNPV);
    	}
    	
    	/**
    	 * Get the positive predictive value of this classifier.
    	 * tp/(tp+fp)
    	 *
    	 */
    	public double getPPV(){
    		return ppv.doubleValue();
    	}
    	
    	/**
    	 * Get the negative predictive value of this classifier.
    	 * tn/(tn+fn)
    	 *
    	 */
    	public double getNPV(){
    		return npv.doubleValue();
    	}
    	
    	/**
    	 * Return the name of this classifier
    	 * @return
    	 */
    	public String getName(){
    		return name;
    	}
    	
    	/**
    	 * Get the modified positive predictive value of this classifier.
    	 * tp/(tp+fp) - baseline Positive
    	 *
    	 */
    	public double getModPPV(){
    		return modPPV.doubleValue();
    	}
    	
    	/**
    	 * Get the modified negative predictive value of this classifier.
    	 * tn/(tn+fn) - baseline Negative
    	 *
    	 */
    	public double getModNPV(){
    		return modNPV.doubleValue();
    	}
    	
    	public String toString(){
    		String strVal="name: "+name+"\n\tTP: "+tp+"\n\tFP: "+fp+"\n\tFN: "+fn+"\n\tTN: "+tn;
    		strVal=strVal+"\n\tPPV: "+ppv+"\n\tNPV: "+npv;
    		strVal=strVal+"\n\tBase Positive: "+basePositive+"\n\tBase Negative: "+baseNegative;
    		strVal=strVal+"\n\tModified PPV: "+modPPV+"\n\tModified NPV: "+modNPV;
    		return strVal;
    	}
    	
    	/**
    	 * Calculate P(C|A) as TP/(TP+FN) if pred active or FN/(TP+FN) if pred inactive
    	 * @param prediction
    	 * @return
    	 */
    	public double calcP_CgA(String prediction){
    		double P_CgA=Double.NaN;
    		
    		double dTP=tp.doubleValue();
    		double dFP=fp.doubleValue();
    		double dFN=fn.doubleValue();
    		double dTN=tn.doubleValue();
    		
    		if (prediction.equalsIgnoreCase("active")){
     			P_CgA=dTP/(dTP+dFN);
    		} else {
     			P_CgA=dFN/(dTP+dFN);
    		}
    		
    		return P_CgA;
    	}
    	
    	/**
    	 * Calculate P(C|D) as TP/(TP+FN)*prior + FP/(FP+TN) * (1-prior) if active, or 
    	 *                     FN/(TP+FN)*prior + TN/(FP+TN) * (1-prior) if inactive
    	 * @param prediction
    	 * @param prior
    	 * @return
    	 */
    	public double calcP_CgD(String prediction, double prior){
    		double P_CgD=Double.NaN;
    		
    		double dTP=tp.doubleValue();
    		double dFP=fp.doubleValue();
    		double dFN=fn.doubleValue();
    		double dTN=tn.doubleValue();
    		
    		if (prediction.equalsIgnoreCase("active")){
     			P_CgD=prior * dTP/(dTP+dFN) + (1-prior)*dFP/(dFP+dTN);
    		} else {
    			P_CgD=prior * dFN/(dTP+dFN) + (1-prior)*dTN/(dFP+dTN);
    		}
    		
    		return P_CgD;
    	}

		/* 
		 * Sort based on tp/(tp+fn) value
		 */
		public int compareTo(Object arg0) {
			ComponentClassifier other = (ComponentClassifier)arg0;
			
			double dTP=this.tp.doubleValue();
    		double dFP=this.fp.doubleValue();
    		double dFN=this.fn.doubleValue();
    		double dTN=this.tn.doubleValue();

			double oTP=other.tp.doubleValue();
    		double oFP=other.fp.doubleValue();
    		double oFN=other.fn.doubleValue();
    		double oTN=other.tn.doubleValue();

    		double compvalue = dTP/(dTP+dFN) - oTP/(oTP+oFN);
		
			return (int)compvalue;
		}
    	
		/**
		 * Sensitivity = tp / (tp + fn)
		 * @return
		 */
		public double getSensitivity(){
			double dTP=this.tp.doubleValue();
    		double dFP=this.fp.doubleValue();
    		double dFN=this.fn.doubleValue();
    		double dTN=this.tn.doubleValue();
    		
    		return (dTP/(dTP+dFN));
		}
		
    } //End component classifier inner class
	
	
	private Hashtable components; //The component classifier objects stored here
	private Double maxModPPV;  //The highest possible modified PPV.  Used to normalize
	private Double maxModNPV;  //The highest possible modified NPV.  Used to normalize
	
		
	/**
	 * Make a classification using a bayes classifier trained 
	 * with components using tp,fp,fn,tp data 
	 * @param trainName
	 * @param testName
	 */
	public Bayes(String trainName, String testName, String outDir, int method){
		readCCfile (trainName);
		//printComponents();	//debug
		//System.out.println("maxModPPV: "+maxModPPV); //debug
		//System.out.println("maxModNPV: "+maxModNPV); //debug
		Predict(testName,outDir, method);
		
	}
	
	/**
	 * Predict the mutants, assume that readCCfile has already been called
	 * @param testName
	 * @param outDir
	 * @param method 1=sum, 2=Bayes, 3=sorted Bayes
	 */
	
	private void Predict(String testName, String outDir, int method){
		ArrayList testFileHeader=new ArrayList();
		

		try{
			FileReader inFileReader = new FileReader(testName);
			BufferedReader inFile = new BufferedReader(inFileReader);
			
			//Read in the header
			String line = inFile.readLine();
			StringTokenizer tok = new StringTokenizer(line);
			while(tok.hasMoreTokens()){
				testFileHeader.add(tok.nextToken());
			}
			line = inFile.readLine();
			
			//Predict each line...  Write results to .result file
			while (line != null){
				double predVal=0;
				String pred="?";
				
				tok = new StringTokenizer(line);
				String name=tok.nextToken();
				
				String truthStr = getTruth(line,testFileHeader);
				
				//Make the predictions:  1)Sum, 2)Bayes, 3)Presorted Bayes	
				double decisionBoundary=0;
				if (method ==1){
					predVal = predUsingSum(line, testFileHeader);
					decisionBoundary=0;
				} else if (method == 2){
					predVal = predUsingBayes(line, testFileHeader);
					decisionBoundary=0.5;
				} else {
					predVal = predUsingSortBayes(line, testFileHeader);
					decisionBoundary=0.5;
				}
				
				if (predVal>=decisionBoundary){
					pred="active";
				} else {
					pred="inactive";
				}
				writeResult(outDir,name,pred,truthStr,predVal);
				line = inFile.readLine();
			}
			
		} catch (Exception e){
			e.printStackTrace();
		}

	}
	
	private String getTruth(String predictions,ArrayList testFileHeader){
		StringTokenizer tok = new StringTokenizer(predictions);
		String name=tok.nextToken();

		String truthStr="?";
			
		int compIdx=1; //Used to index testFileHeader
		while(tok.hasMoreTokens()){
			String curComponent=(String)(testFileHeader.get(compIdx));
			String curPred=tok.nextToken();
			if (curComponent.equalsIgnoreCase("truth")){
				int commaIdx=curPred.length() - 1;
				truthStr=curPred.substring(0,commaIdx);
			}
			compIdx++;
		}
		
		return truthStr;
	}
	
	/**
	 * Sort the bayes in increasing order of sensitivity
	 * @param predictions
	 * @param testFileHeader
	 * @return
	 */
	private double predUsingSortBayes(String predictions,ArrayList testFileHeader){
		//I need to sort testFileHeader using the component classifiers 
		ArrayList compArray=new ArrayList();
		Hashtable headerHash= new Hashtable(); // Used to reference testFileHeader vals by val
		ArrayList newFileHeader= new ArrayList();
		
		String nameTag=(String)(testFileHeader.get(0));
		String truthTag=null;
		int truthIdx=-1;
		for (int i=1; i< testFileHeader.size(); i++){
			String curComponent=(String)(testFileHeader.get(i));
			if (curComponent.equalsIgnoreCase("truth")){
				truthTag=curComponent;
				truthIdx=i;
			} else {
				ComponentClassifier curComp=(ComponentClassifier)(components.get(curComponent));
				Double sensitivity = new Double(curComp.getSensitivity());
				compArray.add(sensitivity);
				headerHash.put(sensitivity, curComponent);
			}
		}
		
		Collections.sort(compArray);
		
		newFileHeader.add(nameTag);
		for (int i=0; i< compArray.size(); i++){
			Double curVal=(Double)(compArray.get(i));
			String curHeader=(String)(headerHash.get(curVal));
			newFileHeader.add(curHeader);
		}
		
		if (truthTag != null){
			newFileHeader.add(truthIdx,truthTag);
		}
		
		return predUsingBayes(predictions,newFileHeader);
	}
	
	/**
	 * P(A|D) = P(A) * P(D|A) / P(D)
	 * @param predictions
	 * @param testFileHeader
	 * @return P(A|D)
	 */
	private double predUsingBayes(String predictions,ArrayList testFileHeader){	
		double prior = 0.5;
		
		StringTokenizer tok = new StringTokenizer(predictions);
		String name=tok.nextToken();

		int compIdx=1; //Used to index testFileHeader
		while(tok.hasMoreTokens()){
			String curComponent=(String)(testFileHeader.get(compIdx));
			String curPred=tok.nextToken();
						
			if (curComponent.equalsIgnoreCase("truth")){
				int commaIdx=curPred.length() - 1;
				String truthStr=curPred.substring(0,commaIdx);
			} else {
				ComponentClassifier curComp=(ComponentClassifier)(components.get(curComponent));
				prior = prior * curComp.calcP_CgA(curPred) / curComp.calcP_CgD(curPred,prior);
			}
			compIdx++;
		}		
		return prior;
	}
	
	/**
	 * Prediction = sum [P(A|D)] - sum [P(I|D)] 
	 * @param predictions
	 * @param testFileHeader
	 * @param truthStr
	 * @return  x > 0 for active, x < 0 for inactive
	 */    
	private double predUsingSum(String predictions,ArrayList testFileHeader){
		double predVal=0;
		
		double ppvNorm=1/(maxModPPV.doubleValue());
		double npvNorm=1/(maxModNPV.doubleValue());
		
		StringTokenizer tok = new StringTokenizer(predictions);
		String name=tok.nextToken();

		
		int compIdx=1; //Used to index testFileHeader
		while(tok.hasMoreTokens()){
			String curComponent=(String)(testFileHeader.get(compIdx));
			String curPred=tok.nextToken();
			
			if (curComponent.equalsIgnoreCase("truth")){
				int commaIdx=curPred.length() - 1;
				String truthStr=curPred.substring(0,commaIdx);
			} else {
				ComponentClassifier curComp=(ComponentClassifier)(components.get(curComponent));
				double pv;
				if (curPred.equalsIgnoreCase("active")){
					pv=curComp.getModPPV() * ppvNorm;
				} else {
					pv=-1*(curComp.getModNPV()*npvNorm);
				}
				predVal=predVal+pv;
			}
			compIdx++;
		}
		return predVal;
	}
	
	private void writeResult(String dir, String name, String pred, String truth, double prob){
		try{
			FileWriter outFile = new FileWriter(dir+"/"+name+".result");
			PrintWriter outArff = new PrintWriter(outFile);
			outArff.println("PRED:\t"+pred);
			outArff.println("TRUTH:\t"+truth);
			outArff.println("P:\t"+prob);
			outArff.close();
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Try to create the component classifier objects 
	 * @param trainName
	 */
	private void readCCfile (String trainName){
		components= new Hashtable();
		
		try{
			FileReader inFileReader = new FileReader(trainName);
			BufferedReader inFile = new BufferedReader(inFileReader);
			
			String line = inFile.readLine();
			while (line!=null){
				if (line.startsWith("#")){
					StringTokenizer tok = new StringTokenizer(line);
					
					String name=tok.nextToken();
					name=name.substring(1);
					
					String stp=tok.nextToken();
					int tp = Integer.parseInt(stp);
					
					String sfp=tok.nextToken();
					int fp = Integer.parseInt(sfp);
					
					String sfn=tok.nextToken();
					int fn = Integer.parseInt(sfn);

					String stn=tok.nextToken();
					int tn = Integer.parseInt(stn);
					
					ComponentClassifier curComponent=new ComponentClassifier(name,tp,fp,fn,tn);
					components.put(name,curComponent);
				}
				line=inFile.readLine();
			}
			
			findMaxPV();
			
		} catch (Exception e){
			e.printStackTrace();
		}

	}
	
	private void findMaxPV(){
		maxModPPV = new Double(0);
		maxModNPV = new Double(0);
		
		//Determine the highest modified PPV and NPV
		Enumeration e = components.elements();
		while(e.hasMoreElements()){
			ComponentClassifier curComponent=(ComponentClassifier)(e.nextElement());
			double curModPPV=curComponent.getModPPV();
			
			if (curModPPV > maxModPPV.doubleValue()){
				maxModPPV = new Double(curModPPV);
			}

			double curModNPV=curComponent.getModNPV();
			if (curModNPV > maxModNPV.doubleValue()){
				maxModNPV = new Double(curModNPV);
			}
		}
	}
	
	
	/**
	 * Print out the current component classifiers
	 *
	 */
	private void printComponents(){
		Enumeration e = components.elements();
		while(e.hasMoreElements()){
			ComponentClassifier curComponent=(ComponentClassifier)(e.nextElement());
			System.out.println(curComponent.toString());
		}		
	}
	
	/**
	 * Provide command line interface
	 * @param args
	 */
	public static void main (String args []){
		if (args.length != 4){
			System.out.println("Use a naive bayes approach to predict a classifier behavior");
			System.out.println("\tBayes <trainFile> <testFile> <outDir> <method>");
			System.out.println("\t<method> = 1, Sum method");
			System.out.println("\t<method> = 2, Bayes method");
			System.out.println("\t<method> = 3, Sorted Bayes method");
			System.out.println("ex: Bayes trainFile.cc testFile.NB ./results 1");
		} else {
			Bayes myBayes = new Bayes(args[0], args[1], args[2], Integer.parseInt(args[3]));
		}
	}
}


/*
 * Version 2.0  Sam Danziger 4/27/05 15:01
 *   Modified to accept three different types of bayes algorithms
 */