import java.io.*;
import java.util.*;
/**
 *
 * @author sdanzige
 *
 * Read in all of the .result files in a directory (or in a list) and calculate
 * the 2x2 correlation coefficient
 */

class CorrelateResults {
	int tp;
	int tn;
	int fp;
	int fn;

	ArrayList results;

	/**
	 * Read in all .result files in a directory that also appear in listName
	 * @param dirName
	 * @param listName
	 */
	public CorrelateResults(String dirName, String listName){
		try{
			FileReader inFile = new FileReader(listName);
			BufferedReader in = new BufferedReader(inFile);

			ArrayList fileList = new ArrayList();
			String line = in.readLine();
			while(line != null){
				StringTokenizer tok = new StringTokenizer(line);
				String lne=tok.nextToken();
				StringTokenizer tok2 = new StringTokenizer(lne,".");
				line = tok2.nextToken();
				fileList.add(line+".result");
				//System.out.println(line);
				line = in.readLine();
			}
			in.close();
			recordResults(dirName, fileList);

		}catch (Exception e){
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
	}

	/**
	 * Read in all .result files in a directory
	 * @param dirName
	 */
	public CorrelateResults(String dirName){
		try{
			File myDir = new File(dirName);
			String [] files = myDir.list(new ResultFileFilter());

			ArrayList fileList = new ArrayList();
			for (int i = 0; i<files.length; i++){
				fileList.add(files[i]);
				//System.out.println(files[i]);
			}

			recordResults(dirName, fileList);
		}catch (Exception e){
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
	}

	/**
	 * Fille the results ArrayList with instances of ReadResults
	 * @param dirName
	 * @param fileList
	 */
	private void recordResults(String dirName, ArrayList fileList){
		tp=0;
		tn=0;
		fp=0;
		fn=0;

		results = new ArrayList();
		for (int i=0; i<fileList.size(); i++){
			String curFile = (String)(fileList.get(i));
			ReadResult tempRR = new ReadResult(dirName + "/" + curFile);
			parseResult(tempRR, curFile);
			results.add(tempRR);
		}
	}

	/**
	 * Record the confusion value of result
	 * @param result
	 */
	private void parseResult(ReadResult result, String curFile){
		String confusion = result.getConfusion("active");
		if (confusion == null){
			System.err.println(curFile+" is corrupt?");
		} else if (confusion.equalsIgnoreCase("TP")){
			tp++;
		} else if (confusion.equalsIgnoreCase("FP")){
			fp++;
		} else if (confusion.equalsIgnoreCase("FN")){
			fn++;
		} else if (confusion.equalsIgnoreCase("TN")){
			tn++;
		}
	}

	/**
	 * Calculate the 2x2 correlation coeffience given a confusion matrix
	 * @return
	 */
	public double get2x2(){
		//System.out.println("tp = "+tp);
		//System.out.println("tn = "+tn);
		//System.out.println("fp = "+fp);
		//System.out.println("fp = "+fp);

		double num1 = (double)(tp*tn);
		//System.out.println("num1 = "+num1);
		double num2 = (double)(fp*fn);
		//System.out.println("num2 = "+num2);
		double num = num1-num2;
		//System.out.println("num = "+num);
		double den1 = tp+fp;
		//System.out.println("den1 = "+den1);
		double den2 = tp+fn;
		//System.out.println("den2 = "+den2);
		double den3 = tn+fp;
		//System.out.println("den3 = "+den3);
		double den4 = tn+fn;
		//System.out.println("den4 = "+den4);
		double den = den1*den2*den3*den4;
		//System.out.println("den5 = "+den);
		den = Math.sqrt(den);
		//System.out.println("den6 = "+den);


		if ((num == 0) && (den == 0)){
			return 0;
		} else {
			double cc = num/den;
			return cc;
		}
	}
	
	/**
	 * Calculate the accuracy: TP+TN / (TP+TN+FP+FN)
	 * @return
	 */
	public double getAccuracy () {
		double num = (double)(tp+tn);
		double den = (double)(tp+tn+fp+fn);
		return (num/den);
	}
	
	/**
	 * Return the accruacy as a formated percentage with n places after the decimal
	 * @return
	 */
	public String getPrettyAccuracy (int n) {
		double acc = this.getAccuracy();
		//System.out.println("acc="+acc);
		acc = acc*Math.pow(10,n+2);
		//System.out.println("acc="+acc);
		long tempValue = Math.round(acc);
		//System.out.println("tempValue="+tempValue);
		double tempValue2 = (double)tempValue;
		//System.out.println("tempValue2="+tempValue2);
		tempValue2 = tempValue2/Math.pow(10,n);
		String percentage = (new Double(tempValue2)).toString()+" %";
		return percentage;
	}

	/**
	 * Return the confusion matrix
	 * @return
	 */
	public String getConfusion(){
		String confusion = "tp = "+tp+"\nfp = "+fp+"\nfn = "+fn+"\ntn = "+tn;
		return confusion;
	}
	
	/**
	 * Return the confusion matrix in 1 line format for bayes parseing
	 * @return
	 */
	public String getOneLine( String name ){
		String confusion = "#"+name+"\t"+tp+"\t"+fp+"\t"+fn+"\t"+tn;
		return confusion;
	}

	/**
	 * Command line
	 * 	Read in a results directory and display it's 2x2 correlation coefficient and confusion matrix
	 *	CoorelateResults <dirName> <listName - optional>
	 *	Ex: CorrelateResults ./2d
	 *	  Include listname to just read files contained in <listName>
	 *
	 * @param args
	 */
	public static void main (String args []) {

		if ((args.length < 1)&&(args.length > 3)){
			System.out.println("Read in a results directory and display it's 2x2 coorelation coefficient and confusion matrix" );
			System.out.println("CorrelateResults <dirName> <featureName - optional> <listName - optional>" );
			System.out.println("Ex: CorrelateResults ./results 2d" );
			System.out.println("\t Include listname to just read files contained in <listName>" );
		} else {
			CorrelateResults myCR = null;
			if (args.length <= 2){
				myCR = new CorrelateResults(args[0]);
			} else {
				myCR = new CorrelateResults(args[0],args[2]);
			}
			
                        String featureName;
                        if (args.length < 2 ){
                                featureName="Unnamed";
                        } else {
                                featureName=args[1];
                        } 

			System.out.println(myCR.getOneLine(featureName));
			System.out.println(myCR.getConfusion());
			System.out.println("Accuracy = " + myCR.getAccuracy() + " = " +myCR.getPrettyAccuracy(2));
			System.out.println("2x2 Correlation Coefficient = "+ myCR.get2x2());
		}
	}
}
