import java.io.*;
import java.util.*;

/**
 * 
 * @author Sam Danziger
 *
 * Extract information from a .pdb file to give the relative postions of the:
 *  1) 4 ZN binding AAs
 *  2) 3 DNA center AAs
 */
class ExtractFeaturesFromPdb{
	/**
	 * 
	 * @author Sam Danziger
	 *
	 * Store the X, Y , and Z coordinates of an atom
	 */
	class AtomInfo {
		public String Name;
		public Double X;
		public Double Y;
		public Double Z;
		
		public AtomInfo(String name, Double x, Double y, Double z){
			Name = name;
			X = x;
			Y = y;
			Z = z;
		}
		
		public AtomInfo(Double x, Double y, Double z){
			Name = new String("Unnamed");
			X = x;
			Y = y;
			Z = z;
		}
		
		/**
		 * Return the distance between two AtomInfos
		 * @param refAI
		 * @return
		 */
		public double calcDistance(AtomInfo refAI){
			double dist=0;
			double x1 = X.doubleValue();
			double y1 = Y.doubleValue();
			double z1 = Z.doubleValue();
			
			double x2 = refAI.X.doubleValue();
			double y2 = refAI.Y.doubleValue();
			double z2 = refAI.Z.doubleValue();
			
			dist = Math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
			
			return dist; 
		}
	}
	
	
	public ExtractFeaturesFromPdb(String mutName, String pdbName, String timeStep){
		/*Create the list of interesting atoms*/
		ArrayList aaNums = new ArrayList();
		aaNums.add(new Integer(81)); //Zn1
		aaNums.add(new Integer(84)); //Zn2
		aaNums.add(new Integer(143)); //Zn3
		aaNums.add(new Integer(147)); //Zn4
		
		ArrayList aaNames = new ArrayList();
		aaNames.add("Zn1"); //Zn1
		aaNames.add("Zn2"); //Zn2
		aaNames.add("Zn3"); //Zn3
		aaNames.add("Zn4"); //Zn4
		
		/*Read in the Alpha Carbon positions*/
		ArrayList acList = getAlphaCarbons(pdbName, aaNums);
			
		/*Determine the Distances*/
		ArrayList attrNames=new ArrayList();
		ArrayList data=new ArrayList();
		
		//Process the 4 Zn atoms
		double znSum=0;
		double numZn=6;
		for(int i=0; i<4; i++){
			AtomInfo refAI = (AtomInfo)(acList.get(i));
			for (int j=i; j<4; j++){
				if (i!=j){
					AtomInfo compAI = (AtomInfo)(acList.get(j));
					double znDist = refAI.calcDistance(compAI);
					attrNames.add(new String(aaNames.get(i)+"_"+aaNames.get(j)+"_"+timeStep));
					data.add(new Double(znDist));
					znSum=znSum+znDist;
				}
			}
		}
		attrNames.add(new String("Zn_Mean_"+timeStep));
		data.add(new Double(znSum/numZn));
		
		//Process the 3 DNA atoms
		aaNums = new ArrayList();
		aaNums.add(new Integer(187)); //DNA1
		aaNums.add(new Integer(153)); //DNA2
		aaNums.add(new Integer(25)); //DNA3

		aaNames = new ArrayList();
		aaNames.add("DNA1"); //DNA1
		aaNames.add("DNA2"); //DNA2
		aaNames.add("DNA3"); //DNA3

		double dnaSum=0;
		double numDNA=3;
		for(int i=0; i<3; i++){
			AtomInfo refAI = (AtomInfo)(acList.get(i));
			for (int j=i; j<3; j++){
				if (i!=j){
					AtomInfo compAI = (AtomInfo)(acList.get(j));
					double dnaDist = refAI.calcDistance(compAI);
					attrNames.add(new String(aaNames.get(i)+"_"+aaNames.get(j)+"_"+timeStep));
					data.add(new Double(dnaDist));
					dnaSum=dnaSum+dnaDist;
				}
			}
		}
		attrNames.add(new String("DNA_Mean_"+timeStep));
		data.add(new Double(dnaSum/numDNA));
		
		printArff(pdbName, attrNames, "NUMERIC", data, mutName);
	}
	
	public void printArff(String relation, ArrayList attributes, String attrType, ArrayList data, String mutName){
		System.out.println("@RELATION\t"+relation);
		for (int i=0; i<attributes.size(); i++){
			System.out.println("@ATTRIBUTE\t"+attributes.get(i)+"\t"+attrType);
		}
		System.out.println("@DATA");
		System.out.println("%"+mutName);
		for (int i=0; i<data.size(); i++){
			System.out.print(data.get(i)+",");
		}
	}
	
	/**
	 * Given the list of Amino Acid numbers, return the Alpha Carbon positions
	 * @param aaNums
	 * @return
	 */
	private ArrayList getAlphaCarbons(String pdbName, ArrayList aaNums){
		ArrayList alphaCarbons = new ArrayList();
		//Arrays.sort(aaNums);
				
		try{
			FileReader inFile = new FileReader(pdbName);
			BufferedReader in = new BufferedReader(inFile);
			
			String line = in.readLine();
			while (line != null){
				StringTokenizer tok = new StringTokenizer(line);
				String curTok = tok.nextToken();
				String atomNum = tok.nextToken();
				String atomName = tok.nextToken();
				String aaName = tok.nextToken();
				String strAANum = tok.nextToken();
				int intAANum = Integer.parseInt(strAANum);
				Integer aaNum = new Integer(intAANum);
				
				//Check to see if this line contains alpha carbon information for an interesting aa
				if (aaNums.contains(aaNum) && atomName.equalsIgnoreCase("CA")){
					Double x = Double.valueOf(tok.nextToken());
					Double y = Double.valueOf(tok.nextToken());
					Double z = Double.valueOf(tok.nextToken());
					
					alphaCarbons.add(new AtomInfo(aaName+aaNum.toString(), x, y, z));
				}
				line = in.readLine();
			}			
		
			in.close();
		} catch (Exception e){
			e.printStackTrace();
		}
		
		return alphaCarbons;
	}

	/**
	 * Extract information from a .pdb file to give the relative postions of the:
 	 *  1) 4 ZN binding AAs
 	 *  2) 3 DNA center AAs
	 * @param args
	 */
	public static void main (String args []){
		if (args.length != 3){
			System.out.println("Extract information from a .pdb file to give the relative postions of the:\n\t1) 4 ZN binding AAs\n\t2) 3 DNA center AAs\nReturn the results in .arff format");
			System.out.println("\tExtractFeaturesFromPdb <mutName> <pdbFile> <indexNum> ");
			System.out.println("ex: ExtractFeaturesFromPdb r273l 1.pdb 1 > 4dfeatures.txt:");
		} else {
			ExtractFeaturesFromPdb myExtractFeaturesFromPdb = new ExtractFeaturesFromPdb(args[0], args[1], args[2]);
		}
	}


}

