import java.io.*;
import java.util.*;

/**
 * 
 * @author Sam Danziger
 *
 * Remove the Zinc Binding and DNA contact mutants from a .arff file
 * and write them as "inactive" in a .result file
 */
class RemoveZincDNA{
	ArrayList contactList; //The residue number for all Zinc and DNA contact
	
	public RemoveZincDNA(String arffName){
		fillContactList();
		try {
			FileReader inFile = new FileReader(arffName);
			BufferedReader in = new BufferedReader(inFile);
			
			String outName = addNoContact(arffName);
			PrintWriter out = new PrintWriter( new FileWriter(outName));
			
			//Header
			String line = in.readLine();
			while (!(line.toLowerCase()).startsWith("@data")){
				out.println(line);
				line = in.readLine();
			}
			out.println(line);
			
			//Data
			line = in.readLine();
			while (line != null){
				if (isContactOrBinding(line.toLowerCase())){
					String mutName = getMutName(line);
					line = in.readLine();
					String activity = getTrueActivity(line);
					writeResult(mutName,activity);
				} else {
					out.println(line);
					line = in.readLine();
					out.println(line);
				}
				line = in.readLine();
			}
			out.close();
			in.close();
		} catch (Exception e){
			System.out.println(e.getMessage());
			e.printStackTrace();			
		}
	}
	
	/**
	 * ZN+2
	 * 171-184
	 * 233-247
	 *
	 * DNA Binding:
	 * H2: 272-286
	 * L1: 117-123
	 * L3: 245-251
	 */
	private void fillContactList(){
		contactList = new ArrayList();
		
		int min = 171;
		int max = 184;
		for (int i = min; i <= max; i++){
			contactList.add(new Integer(i));
		}

	    min = 233;
		max = 247;
		for (int i = min; i <= max; i++){
			contactList.add(new Integer(i));
		}
		
		min = 272;
		max = 286;
		for (int i = min; i <= max; i++){
			contactList.add(new Integer(i));
		}
		
		min = 117;
		max = 123;
		for (int i = min; i <= max; i++){
			contactList.add(new Integer(i));
		}
		
		min = 245;
		max = 251;
		for (int i = min; i <= max; i++){
			contactList.add(new Integer(i));
		}		
	}
	
	/**
	 * Put the "noContact" into the fileName 
	 * @param fileName
	 * @return
	 */
	private String addNoContact (String fileName){
		String newName = null;
		
		int idx = fileName.lastIndexOf("/");
		if (idx == -1){
			newName = "noContact."+fileName;
		} else {
			newName = fileName.subSequence(0,idx+1)+"noContact."+fileName.subSequence(idx+1,fileName.length());
		}
		return newName;
	}
	
	/**
	 * Returns true if line indicates a contact or DNAbinding mutation
	 * @param line
	 * @return
	 */
	private boolean isContactOrBinding(String line){
		for(int i = 0; i<contactList.size(); i++){
		   	Integer curLoc = (Integer)(contactList.get(i));
		   	String strCurLoc = curLoc.toString();
		   	if (line.indexOf(strCurLoc)!=-1){
		   		return true;
		   	}
		}
		return false;
	}
	
	/**
	 * Given the mutants Name, and it's true activity, 
	 * write the results file
	 * @param mutName
	 * @param activity
	 */
	private void writeResult (String mutName, String activity){
		try {
			String outName = "./results/"+mutName+".result";
			PrintWriter out = new PrintWriter( new FileWriter(outName));
			out.println("PRED:\tinactive");
			out.println("TRUTH:\t"+activity);
			out.close();
		} catch (Exception e){
			System.out.println(e.getMessage());
			e.printStackTrace();			
		}
		
	}
	
	/**
	 *Parse a comment line to get the mutant name 
	 * @param line
	 */ 
	private String getMutName (String line){
		StringTokenizer tok = new StringTokenizer(line,"%");
		String newMut = tok.nextToken();
		tok = new StringTokenizer(newMut);
		newMut = tok.nextToken();
		return newMut;
	}
	
	/**
	 * Parse a data line to get the last value (the activity)
	 * @param line
	 * @return
	 */
	private String getTrueActivity(String line){
		int lastIdx = line.lastIndexOf(",");
		int nlIdx = (line.substring(0,lastIdx)).lastIndexOf(",");
		String activity = line.substring(nlIdx+1,lastIdx);
		return activity;
	}
	
	/**
	 * Remove the DNA contact and Zinc Binding domain mutants
	 * from a .arff file and save it as .result files
	 * RemoveZincDNA <arffName>
	 * RemoveZincDNA 3d.arff
	 * 
	 * @param args
	 */
	public static void main (String args []) {
		if (args.length != 1){
			System.out.println("Filter repeats out of a .arff file" );
			System.out.println("RemoveZincDNA <arffName>" );
			System.out.println("RemoveZincDNA 3d.arff" );
		} else {
			String arffName = args[0];
			RemoveZincDNA myRemoveZincDNA = new RemoveZincDNA(arffName);
			System.out.println("Completed Execution: RemoveZincDNA "+arffName);
		}
	}
}