import java.io.*;
import java.util.*;

class ResultFileFilter implements FilenameFilter {
		public boolean accept(File dir, String file) {
				return (file.toLowerCase()).endsWith(".result");
		}
}


/**
 *
 * @author Sam Danziger
 *
 * Read in the results file from the 2d/3d/4d and record the results
 * in an .arff file
 */
class Result2Arff {
		ArrayList dirNames;      //All of the directories 2d, 3d, etc.
		ArrayList mutNamesLists; //The lists of all mutant files in each dir
		ArrayList isectList;
		ArrayList exclList;
		ArrayList unionList;

		/**
		 * Convert the results file to a .arff file
		 * @param dirList  2d,3d,4d,etc
		 * @param outName  the name of the output file
		 */
		public Result2Arff(String dirListName, String outName, String startDir, String resultDir){
				isectList = new ArrayList();
				exclList = new ArrayList();
				unionList = new ArrayList();
				mutNamesLists = new ArrayList();

				readDirList(dirListName);
				readMutantNames(startDir, resultDir);
				makeExclusive();
				writeArff(outName+".noQuestion.arff",startDir,resultDir);
				writeArffwQuestion(outName+".arff",startDir,resultDir);
				writeList(exclList, outName+".excl");
			    writeNB(outName+".NB", startDir, resultDir);
		}

	    /**
	     * Write the output to a file for the Naive Bayes classifier
	     * @param arffName
	     * @param startDir
	     */
	    private void writeNB(String arffName, String startDir, String resultDir){
			try{
					//Write Header
					PrintWriter output = new PrintWriter(new FileWriter(arffName));
					
					output.print("#NAME");
					int numAttrib = dirNames.size();
					for (int i=0; i<numAttrib; i++){
							output.print("\t"+(String)(dirNames.get(i)));
					}
					output.print("\ttruth\n");
					//output.print("\n");

					//Process each results file
					int numMuts = unionList.size();
					ReadResult resReader = null;
					String outFileName;
					numMuts = unionList.size();
					resReader = null;
					String trueActivity = null;
					for (int i=0; i<numMuts; i++){
							output.print(removeExt((String)(unionList.get(i))));
							for (int j=0; j<numAttrib; j++){
									String inFileName = (String)(startDir+"/"+dirNames.get(j))+"/"+resultDir+"/"+(String)(unionList.get(i));
									resReader = new ReadResult(inFileName);
									if (resReader.getPred() == null) {
											output.print("\t?");
									} else {
											output.print("\t"+resReader.getPred());
									}
										
									if (resReader.getTruth()!=null){
										trueActivity = resReader.getTruth();
									} else {
										trueActivity = "?";
									}
							}
							//Write the actualy activity
							output.print("\t"+trueActivity+",\n");
					    	//output.print("\n");
					}
					output.close();

			}catch (Exception e){
					System.out.println(e.getMessage());
					e.printStackTrace();
			}
	    }
	    
	    		
		/**
		 * Write all elements of any collection to a file
		 * @param list
		 * @param fileName
		 */
		private void writeList(Collection list, String fileName){
				try{
						PrintWriter output = new PrintWriter( new FileWriter(fileName));
						Object [] objList = list.toArray();
						for(int i=0;i<objList.length;i++){
								output.println(objList[i]);
						}
						output.close();

				}catch (Exception e){
						System.out.println(e.getMessage());
						e.printStackTrace();
				}
		}

		/**
		 * Read the input file with the directory list into the arraylist
		 * @param dirListName
		 */
		private void readDirList(String dirListName){
				dirNames = new ArrayList();
				try{
						FileReader inFile = new FileReader(dirListName);
						BufferedReader in = new BufferedReader(inFile);

						String line = in.readLine();
						while (line != null){
								dirNames.add(line);
								line = in.readLine();
						}

						in.close();
				}catch (Exception e){
						System.out.println(e.getMessage());
						e.printStackTrace();
				}
		}

		/**
		 * After readDirList, create a mutant list from each directory
		 *
		 */
		private void readMutantNames(String startDir, String resultDir){
				for (int i=0; i<dirNames.size(); i++){
						String curDir = (String)(dirNames.get(i));

						File myDir = new File(startDir + "/" + curDir + "/"+ resultDir+"/");
						System.out.println("Processing" +startDir + "/" + curDir + "/"+ resultDir+"/");
						String [] files = myDir.list(new ResultFileFilter());

			if (files == null){
				System.err.println("Can't Read: "+startDir + "/" + curDir + "/"+ resultDir+"/");
			} else {
								LinkedHashSet curFileList = new LinkedHashSet();
								for(int j=0; j<files.length; j++){
										curFileList.add(files[j]);
								}
								mutNamesLists.add(curFileList);
			}
				}
		}

		/**
		 * Perform set functions to determine the intersected list and the
		 * excluded lists
		 *
		 */
		private void makeExclusive(){
				LinkedHashSet isectSet = new LinkedHashSet();
				LinkedHashSet exclSet = new LinkedHashSet();
				LinkedHashSet unionSet = new LinkedHashSet();

				unionSet = new LinkedHashSet((LinkedHashSet)(mutNamesLists.get(0)));
				isectSet = new LinkedHashSet((LinkedHashSet)(mutNamesLists.get(0)));

				for (int i=1; i<mutNamesLists.size(); i++){
						LinkedHashSet curSet = (LinkedHashSet)(mutNamesLists.get(i));
						unionSet.addAll(curSet); //union
						isectSet.retainAll(curSet); // intersection
				}

				//System.out.println("union = " + unionSet);
				//System.out.println("intersection = " + isectSet);

				exclSet = new LinkedHashSet(unionSet);
				exclSet.removeAll(isectSet);

				isectList = new ArrayList(isectSet);
				exclList = new ArrayList(exclSet);
				unionList = new ArrayList(unionSet);
		}

		/**
		 * Write a .arff file that contains all results files that
		 * exist for all 3
		 * @param arffName
		 */
		private void writeArff(String arffName, String startDir, String resultDir){
				try{
						//Write Header
						PrintWriter output = new PrintWriter(new FileWriter(arffName));
						output.println("%Title: p53 Composite Predictor");
						output.println("%Date: "+(new Date()).toLocaleString());
						output.println("@relation Composite_Predictor");

						int numAttrib = dirNames.size();
						for (int i=0; i<numAttrib; i++){
								output.println("@attribute\t"+(String)(dirNames.get(i))+"\t{active,inactive}");
						}
						output.println("@attribute\t"+"truth"+"\t{active,inactive}");

						//Process each results file
						int numMuts = isectList.size();
						ReadResult resReader = null;
						String outFileName;
						numMuts = isectList.size();
						resReader = null;
						output.println("@data");
						for (int i=0; i<numMuts; i++){
								output.println("%"+removeExt((String)(isectList.get(i))));
								for (int j=0; j<numAttrib; j++){
										String inFileName = (String)(startDir+"/"+dirNames.get(j))+"/"+resultDir+"/"+(String)(isectList.get(i));
										resReader = new ReadResult(inFileName);
										output.print(resReader.getPred()+",");
								}
								output.print(resReader.getTruth()+",\n");
						}
						output.close();

				}catch (Exception e){
						System.out.println(e.getMessage());
						e.printStackTrace();
				}
		}

		/**
		 * Write a .arff file that contains all results files that
		 * exist in any of the 3 with ? in missing spaces
		 * @param arffName
		 */
		private void writeArffwQuestion(String arffName, String startDir, String resultDir){
				try{
						//Write Header
						PrintWriter output = new PrintWriter(new FileWriter(arffName));
						output.println("%Title: p53 Composite Predictor");
						output.println("%Date: "+(new Date()).toLocaleString());
						output.println("@relation Composite_Predictor");

						int numAttrib = dirNames.size();
						for (int i=0; i<numAttrib; i++){
								output.println("@attribute\t"+(String)(dirNames.get(i))+"\t{active,inactive}");
						}
						output.println("@attribute\t"+"truth"+"\t{active,inactive}");

						//Process each results file
						int numMuts = unionList.size();
						ReadResult resReader = null;
						String outFileName;
						numMuts = unionList.size();
						resReader = null;
						String trueActivity = null;
						output.println("@data");
						for (int i=0; i<numMuts; i++){
								output.println("%"+removeExt((String)(unionList.get(i))));
								for (int j=0; j<numAttrib; j++){
										String inFileName = (String)(startDir+"/"+dirNames.get(j))+"/"+resultDir+"/"+(String)(unionList.get(i));
										resReader = new ReadResult(inFileName);
										if (resReader.getPred() == null) {
												output.print("?,");
										} else {
												output.print(resReader.getPred()+",");
										}
										
										if (resReader.getTruth()!=null){
											trueActivity = resReader.getTruth();
										}
								}
								//Write the actualy activity
								output.print(trueActivity+",\n");
						}
						output.close();

				}catch (Exception e){
						System.out.println(e.getMessage());
						e.printStackTrace();
				}
		}

		/**
		 * Remove the extension from a string
		 * @param inStr
		 */
		private String removeExt(String inStr){
				int idx = inStr.lastIndexOf(".");
				String outStr = inStr.substring(0,idx);
				return outStr;
		}

		/**
		 * Combine the results files to make a new .arff file
		 * Result2Arff <dirListName>, <outName>, <startDir -- optional>
		 * Result2Arff dirList composite.arff
		 *
		 * @param args
		 */
		public static void main (String args []) {
				if ((args.length > 4)||(args.length < 2)){
						System.out.println("Combine the results files to make a new .arff file" );
						System.out.println("Result2Arff <dirListName>, <outName>, <startDir -- optional> <resultDirName -- Optional>" );
						System.out.println("Ex: Result2Arff dirList composite.arff . results_xvalid" );
				} else {
						String dirListName = args[0];
						String outName = args[1];
						String startDir;
						if (args.length >= 3) {
							startDir = args[2];
						} else {
							startDir = ".";
						}
						String resultDir;
						if (args.length >= 4) {
							resultDir = args[3];
						} else {
							resultDir = "results";
						}
						
						
						Result2Arff myRes2Arff = new Result2Arff(dirListName,outName,startDir, resultDir);
						System.out.println("Completed Execution: Result2Arff "+dirListName+" "+outName+" "+startDir);
				}
		}

}
